/* $Id: x86_setup.c,v 1.9 2001/02/13 21:12:24 mikpe Exp $
 * Performance-monitoring counters driver.
 * x86-specific kernel-resident code.
 *
 * Copyright (C) 1999-2001  Mikael Pettersson
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <asm/processor.h>
#include <asm/perfctr.h>
#include "compat.h"
#include "x86_compat.h"

void __init perfctr_dodgy_tsc(void)
{
#if defined(CONFIG_PERFCTR_WINCHIP)
	/* Check for Centaur WinChip C6/2/3 with broken TSC */
	if( boot_cpu_data.x86_vendor != X86_VENDOR_CENTAUR )
		return;
	if( boot_cpu_data.x86 != 5 )
		return;
	switch( boot_cpu_data.x86_model ) {
	case 4:	break;		/* WinChip C6 */
	case 8:	break;		/* WinChip 2 */
	case 9: break;		/* WinChip 3 */
	default: return;
	}
	/*
	 * WinChip processors implement the TSC by using the
	 * low 32 bits of the two performance counters.
	 * If the performance counters are reprogrammed, the
	 * RDTSC instruction will yield incorrect results.
	 * (On the C6, a read of the TSC using RDTSC or RDMSR
	 * may even cause the processor to hang in this case.
	 * See erratum I-13 in the C6 data sheet for details.)
	 *
	 * Since the user explicitly requested the performance
	 * counters, we have to disable all uses of the TSC.
	 */
	set_in_cr4(X86_CR4_TSD);
	clr_cap_tsc();
	printk(KERN_INFO "WinChip Time-Stamp Counter disabled\n");
#endif
}

#if PERFCTR_INTERRUPT_SUPPORT
unsigned int apic_lvtpc_irqs[NR_CPUS];

static void perfctr_default_ihandler(unsigned long pc)
{
	++apic_lvtpc_irqs[smp_processor_id()];
}

static perfctr_ihandler_t perfctr_ihandler = perfctr_default_ihandler;

static void __attribute__((unused))
do_perfctr_interrupt(struct pt_regs *regs)
{
	/* XXX: should be rewritten in assembly and inlined below */
	/* XXX: recursive interrupts? delay the ACK, mask LVTPC, or queue? */
	ack_APIC_irq();
	(*perfctr_ihandler)(regs->eip);
}

#define BUILD_PERFCTR_INTERRUPT(x,v) XBUILD_PERFCTR_INTERRUPT(x,v)
#define XBUILD_PERFCTR_INTERRUPT(x,v) \
asmlinkage void x(void); \
__asm__( \
	"\n.text\n\t" \
	__ALIGN_STR "\n\t" \
	".type " SYMBOL_NAME_STR(x) ",@function\n" \
	".globl " SYMBOL_NAME_STR(x) "\n" \
SYMBOL_NAME_STR(x) ":\n\t" \
	"pushl $" #v "\n\t" \
	SAVE_ALL \
	"pushl %esp\n\t" \
	"call " SYMBOL_NAME_STR(do_ ## x) "\n\t" \
	"addl $4,%esp\n\t" \
	"jmp ret_from_intr\n\t" \
	".size " SYMBOL_NAME_STR(x) ",.-" SYMBOL_NAME_STR(x) "\n" \
	".previous\n");

BUILD_PERFCTR_INTERRUPT(perfctr_interrupt,LOCAL_PERFCTR_VECTOR)

void perfctr_cpu_set_ihandler(perfctr_ihandler_t handler)
{
	perfctr_ihandler = handler ? handler : perfctr_default_ihandler;
}
#endif

#ifdef CONFIG_PERFCTR_MODULE
EXPORT_mmu_cr4_features;
EXPORT_cpu_khz;
#ifdef NMI_LOCAL_APIC
EXPORT_SYMBOL(nmi_watchdog);
#endif
#if PERFCTR_INTERRUPT_SUPPORT
EXPORT_SYMBOL(perfctr_cpu_set_ihandler);
#endif
#endif
